# autogenerated by __main__.py do not edit
name_map = {'PYQT_VERSION': 'PyQt6.QtCore',
 'PYQT_VERSION_STR': 'PyQt6.QtCore',
 'QAbstractAnimation': 'PyQt6.QtCore',
 'QAbstractButton': 'PyQt6.QtWidgets',
 'QAbstractEventDispatcher': 'PyQt6.QtCore',
 'QAbstractFileIconProvider': 'PyQt6.QtGui',
 'QAbstractGraphicsShapeItem': 'PyQt6.QtWidgets',
 'QAbstractItemDelegate': 'PyQt6.QtWidgets',
 'QAbstractItemModel': 'PyQt6.QtCore',
 'QAbstractItemView': 'PyQt6.QtWidgets',
 'QAbstractListModel': 'PyQt6.QtCore',
 'QAbstractNativeEventFilter': 'PyQt6.QtCore',
 'QAbstractNetworkCache': 'PyQt6.QtNetwork',
 'QAbstractOpenGLFunctions': 'PyQt6.QtOpenGL',
 'QAbstractPrintDialog': 'PyQt6.QtPrintSupport',
 'QAbstractProxyModel': 'PyQt6.QtCore',
 'QAbstractScrollArea': 'PyQt6.QtWidgets',
 'QAbstractSlider': 'PyQt6.QtWidgets',
 'QAbstractSocket': 'PyQt6.QtNetwork',
 'QAbstractSpinBox': 'PyQt6.QtWidgets',
 'QAbstractTableModel': 'PyQt6.QtCore',
 'QAbstractTextDocumentLayout': 'PyQt6.QtGui',
 'QAction': 'PyQt6.QtGui',
 'QActionEvent': 'PyQt6.QtGui',
 'QActionGroup': 'PyQt6.QtGui',
 'QAnimationGroup': 'PyQt6.QtCore',
 'QApplication': 'PyQt6.QtWidgets',
 'QAudio': 'PyQt6.QtMultimedia',
 'QAudioBuffer': 'PyQt6.QtMultimedia',
 'QAudioDecoder': 'PyQt6.QtMultimedia',
 'QAudioDevice': 'PyQt6.QtMultimedia',
 'QAudioFormat': 'PyQt6.QtMultimedia',
 'QAudioInput': 'PyQt6.QtMultimedia',
 'QAudioOutput': 'PyQt6.QtMultimedia',
 'QAudioSink': 'PyQt6.QtMultimedia',
 'QAudioSource': 'PyQt6.QtMultimedia',
 'QAuthenticator': 'PyQt6.QtNetwork',
 'QBackingStore': 'PyQt6.QtGui',
 'QBasicTimer': 'PyQt6.QtCore',
 'QBitArray': 'PyQt6.QtCore',
 'QBitmap': 'PyQt6.QtGui',
 'QBluetoothPermission': 'PyQt6.QtCore',
 'QBoxLayout': 'PyQt6.QtWidgets',
 'QBrush': 'PyQt6.QtGui',
 'QBuffer': 'PyQt6.QtCore',
 'QButtonGroup': 'PyQt6.QtWidgets',
 'QByteArray': 'PyQt6.QtCore',
 'QByteArrayMatcher': 'PyQt6.QtCore',
 'QCalendar': 'PyQt6.QtCore',
 'QCalendarPermission': 'PyQt6.QtCore',
 'QCalendarWidget': 'PyQt6.QtWidgets',
 'QCamera': 'PyQt6.QtMultimedia',
 'QCameraDevice': 'PyQt6.QtMultimedia',
 'QCameraFormat': 'PyQt6.QtMultimedia',
 'QCameraPermission': 'PyQt6.QtCore',
 'QCapturableWindow': 'PyQt6.QtMultimedia',
 'QCborError': 'PyQt6.QtCore',
 'QCborKnownTags': 'PyQt6.QtCore',
 'QCborSimpleType': 'PyQt6.QtCore',
 'QCborStreamReader': 'PyQt6.QtCore',
 'QCborStreamWriter': 'PyQt6.QtCore',
 'QCheckBox': 'PyQt6.QtWidgets',
 'QChildEvent': 'PyQt6.QtCore',
 'QChildWindowEvent': 'PyQt6.QtGui',
 'QClipboard': 'PyQt6.QtGui',
 'QCloseEvent': 'PyQt6.QtGui',
 'QCollator': 'PyQt6.QtCore',
 'QCollatorSortKey': 'PyQt6.QtCore',
 'QColor': 'PyQt6.QtGui',
 'QColorConstants': 'PyQt6.QtGui',
 'QColorDialog': 'PyQt6.QtWidgets',
 'QColorSpace': 'PyQt6.QtGui',
 'QColorTransform': 'PyQt6.QtGui',
 'QColumnView': 'PyQt6.QtWidgets',
 'QComboBox': 'PyQt6.QtWidgets',
 'QCommandLineOption': 'PyQt6.QtCore',
 'QCommandLineParser': 'PyQt6.QtCore',
 'QCommandLinkButton': 'PyQt6.QtWidgets',
 'QCommonStyle': 'PyQt6.QtWidgets',
 'QCompleter': 'PyQt6.QtWidgets',
 'QConcatenateTablesProxyModel': 'PyQt6.QtCore',
 'QConicalGradient': 'PyQt6.QtGui',
 'QContactsPermission': 'PyQt6.QtCore',
 'QContextMenuEvent': 'PyQt6.QtGui',
 'QCoreApplication': 'PyQt6.QtCore',
 'QCryptographicHash': 'PyQt6.QtCore',
 'QCursor': 'PyQt6.QtGui',
 'QDataStream': 'PyQt6.QtCore',
 'QDataWidgetMapper': 'PyQt6.QtWidgets',
 'QDate': 'PyQt6.QtCore',
 'QDateEdit': 'PyQt6.QtWidgets',
 'QDateTime': 'PyQt6.QtCore',
 'QDateTimeEdit': 'PyQt6.QtWidgets',
 'QDeadlineTimer': 'PyQt6.QtCore',
 'QDesktopServices': 'PyQt6.QtGui',
 'QDial': 'PyQt6.QtWidgets',
 'QDialog': 'PyQt6.QtWidgets',
 'QDialogButtonBox': 'PyQt6.QtWidgets',
 'QDir': 'PyQt6.QtCore',
 'QDirIterator': 'PyQt6.QtCore',
 'QDnsDomainNameRecord': 'PyQt6.QtNetwork',
 'QDnsHostAddressRecord': 'PyQt6.QtNetwork',
 'QDnsLookup': 'PyQt6.QtNetwork',
 'QDnsMailExchangeRecord': 'PyQt6.QtNetwork',
 'QDnsServiceRecord': 'PyQt6.QtNetwork',
 'QDnsTextRecord': 'PyQt6.QtNetwork',
 'QDockWidget': 'PyQt6.QtWidgets',
 'QDoubleSpinBox': 'PyQt6.QtWidgets',
 'QDoubleValidator': 'PyQt6.QtGui',
 'QDrag': 'PyQt6.QtGui',
 'QDragEnterEvent': 'PyQt6.QtGui',
 'QDragLeaveEvent': 'PyQt6.QtGui',
 'QDragMoveEvent': 'PyQt6.QtGui',
 'QDropEvent': 'PyQt6.QtGui',
 'QDynamicPropertyChangeEvent': 'PyQt6.QtCore',
 'QEasingCurve': 'PyQt6.QtCore',
 'QElapsedTimer': 'PyQt6.QtCore',
 'QEnterEvent': 'PyQt6.QtGui',
 'QErrorMessage': 'PyQt6.QtWidgets',
 'QEvent': 'PyQt6.QtCore',
 'QEventLoop': 'PyQt6.QtCore',
 'QEventLoopLocker': 'PyQt6.QtCore',
 'QEventPoint': 'PyQt6.QtGui',
 'QExposeEvent': 'PyQt6.QtGui',
 'QFile': 'PyQt6.QtCore',
 'QFileDevice': 'PyQt6.QtCore',
 'QFileDialog': 'PyQt6.QtWidgets',
 'QFileIconProvider': 'PyQt6.QtWidgets',
 'QFileInfo': 'PyQt6.QtCore',
 'QFileOpenEvent': 'PyQt6.QtGui',
 'QFileSelector': 'PyQt6.QtCore',
 'QFileSystemModel': 'PyQt6.QtGui',
 'QFileSystemWatcher': 'PyQt6.QtCore',
 'QFocusEvent': 'PyQt6.QtGui',
 'QFocusFrame': 'PyQt6.QtWidgets',
 'QFont': 'PyQt6.QtGui',
 'QFontComboBox': 'PyQt6.QtWidgets',
 'QFontDatabase': 'PyQt6.QtGui',
 'QFontDialog': 'PyQt6.QtWidgets',
 'QFontInfo': 'PyQt6.QtGui',
 'QFontMetrics': 'PyQt6.QtGui',
 'QFontMetricsF': 'PyQt6.QtGui',
 'QFormLayout': 'PyQt6.QtWidgets',
 'QFrame': 'PyQt6.QtWidgets',
 'QGenericArgument': 'PyQt6.QtCore',
 'QGenericReturnArgument': 'PyQt6.QtCore',
 'QGesture': 'PyQt6.QtWidgets',
 'QGestureEvent': 'PyQt6.QtWidgets',
 'QGestureRecognizer': 'PyQt6.QtWidgets',
 'QGlyphRun': 'PyQt6.QtGui',
 'QGradient': 'PyQt6.QtGui',
 'QGraphicsAnchor': 'PyQt6.QtWidgets',
 'QGraphicsAnchorLayout': 'PyQt6.QtWidgets',
 'QGraphicsBlurEffect': 'PyQt6.QtWidgets',
 'QGraphicsColorizeEffect': 'PyQt6.QtWidgets',
 'QGraphicsDropShadowEffect': 'PyQt6.QtWidgets',
 'QGraphicsEffect': 'PyQt6.QtWidgets',
 'QGraphicsEllipseItem': 'PyQt6.QtWidgets',
 'QGraphicsGridLayout': 'PyQt6.QtWidgets',
 'QGraphicsItem': 'PyQt6.QtWidgets',
 'QGraphicsItemGroup': 'PyQt6.QtWidgets',
 'QGraphicsLayout': 'PyQt6.QtWidgets',
 'QGraphicsLayoutItem': 'PyQt6.QtWidgets',
 'QGraphicsLineItem': 'PyQt6.QtWidgets',
 'QGraphicsLinearLayout': 'PyQt6.QtWidgets',
 'QGraphicsObject': 'PyQt6.QtWidgets',
 'QGraphicsOpacityEffect': 'PyQt6.QtWidgets',
 'QGraphicsPathItem': 'PyQt6.QtWidgets',
 'QGraphicsPixmapItem': 'PyQt6.QtWidgets',
 'QGraphicsPolygonItem': 'PyQt6.QtWidgets',
 'QGraphicsProxyWidget': 'PyQt6.QtWidgets',
 'QGraphicsRectItem': 'PyQt6.QtWidgets',
 'QGraphicsRotation': 'PyQt6.QtWidgets',
 'QGraphicsScale': 'PyQt6.QtWidgets',
 'QGraphicsScene': 'PyQt6.QtWidgets',
 'QGraphicsSceneContextMenuEvent': 'PyQt6.QtWidgets',
 'QGraphicsSceneDragDropEvent': 'PyQt6.QtWidgets',
 'QGraphicsSceneEvent': 'PyQt6.QtWidgets',
 'QGraphicsSceneHelpEvent': 'PyQt6.QtWidgets',
 'QGraphicsSceneHoverEvent': 'PyQt6.QtWidgets',
 'QGraphicsSceneMouseEvent': 'PyQt6.QtWidgets',
 'QGraphicsSceneMoveEvent': 'PyQt6.QtWidgets',
 'QGraphicsSceneResizeEvent': 'PyQt6.QtWidgets',
 'QGraphicsSceneWheelEvent': 'PyQt6.QtWidgets',
 'QGraphicsSimpleTextItem': 'PyQt6.QtWidgets',
 'QGraphicsTextItem': 'PyQt6.QtWidgets',
 'QGraphicsTransform': 'PyQt6.QtWidgets',
 'QGraphicsVideoItem': 'PyQt6.QtMultimediaWidgets',
 'QGraphicsView': 'PyQt6.QtWidgets',
 'QGraphicsWidget': 'PyQt6.QtWidgets',
 'QGridLayout': 'PyQt6.QtWidgets',
 'QGroupBox': 'PyQt6.QtWidgets',
 'QGuiApplication': 'PyQt6.QtGui',
 'QHBoxLayout': 'PyQt6.QtWidgets',
 'QHeaderView': 'PyQt6.QtWidgets',
 'QHelpEvent': 'PyQt6.QtGui',
 'QHideEvent': 'PyQt6.QtGui',
 'QHostAddress': 'PyQt6.QtNetwork',
 'QHostInfo': 'PyQt6.QtNetwork',
 'QHoverEvent': 'PyQt6.QtGui',
 'QHstsPolicy': 'PyQt6.QtNetwork',
 'QHttp1Configuration': 'PyQt6.QtNetwork',
 'QHttp2Configuration': 'PyQt6.QtNetwork',
 'QHttpHeaders': 'PyQt6.QtNetwork',
 'QHttpMultiPart': 'PyQt6.QtNetwork',
 'QHttpPart': 'PyQt6.QtNetwork',
 'QIODevice': 'PyQt6.QtCore',
 'QIODeviceBase': 'PyQt6.QtCore',
 'QIcon': 'PyQt6.QtGui',
 'QIconDragEvent': 'PyQt6.QtGui',
 'QIconEngine': 'PyQt6.QtGui',
 'QIdentityProxyModel': 'PyQt6.QtCore',
 'QImage': 'PyQt6.QtGui',
 'QImageCapture': 'PyQt6.QtMultimedia',
 'QImageIOHandler': 'PyQt6.QtGui',
 'QImageReader': 'PyQt6.QtGui',
 'QImageWriter': 'PyQt6.QtGui',
 'QInputDevice': 'PyQt6.QtGui',
 'QInputDialog': 'PyQt6.QtWidgets',
 'QInputEvent': 'PyQt6.QtGui',
 'QInputMethod': 'PyQt6.QtGui',
 'QInputMethodEvent': 'PyQt6.QtGui',
 'QInputMethodQueryEvent': 'PyQt6.QtGui',
 'QIntValidator': 'PyQt6.QtGui',
 'QItemDelegate': 'PyQt6.QtWidgets',
 'QItemEditorCreatorBase': 'PyQt6.QtWidgets',
 'QItemEditorFactory': 'PyQt6.QtWidgets',
 'QItemSelection': 'PyQt6.QtCore',
 'QItemSelectionModel': 'PyQt6.QtCore',
 'QItemSelectionRange': 'PyQt6.QtCore',
 'QJsonDocument': 'PyQt6.QtCore',
 'QJsonParseError': 'PyQt6.QtCore',
 'QJsonValue': 'PyQt6.QtCore',
 'QKeyCombination': 'PyQt6.QtCore',
 'QKeyEvent': 'PyQt6.QtGui',
 'QKeySequence': 'PyQt6.QtGui',
 'QKeySequenceEdit': 'PyQt6.QtWidgets',
 'QLCDNumber': 'PyQt6.QtWidgets',
 'QLabel': 'PyQt6.QtWidgets',
 'QLayout': 'PyQt6.QtWidgets',
 'QLayoutItem': 'PyQt6.QtWidgets',
 'QLibrary': 'PyQt6.QtCore',
 'QLibraryInfo': 'PyQt6.QtCore',
 'QLine': 'PyQt6.QtCore',
 'QLineEdit': 'PyQt6.QtWidgets',
 'QLineF': 'PyQt6.QtCore',
 'QLinearGradient': 'PyQt6.QtGui',
 'QListView': 'PyQt6.QtWidgets',
 'QListWidget': 'PyQt6.QtWidgets',
 'QListWidgetItem': 'PyQt6.QtWidgets',
 'QLocalServer': 'PyQt6.QtNetwork',
 'QLocalSocket': 'PyQt6.QtNetwork',
 'QLocale': 'PyQt6.QtCore',
 'QLocationPermission': 'PyQt6.QtCore',
 'QLockFile': 'PyQt6.QtCore',
 'QLoggingCategory': 'PyQt6.QtCore',
 'QMainWindow': 'PyQt6.QtWidgets',
 'QMargins': 'PyQt6.QtCore',
 'QMarginsF': 'PyQt6.QtCore',
 'QMatrix2x2': 'PyQt6.QtGui',
 'QMatrix2x3': 'PyQt6.QtGui',
 'QMatrix2x4': 'PyQt6.QtGui',
 'QMatrix3x2': 'PyQt6.QtGui',
 'QMatrix3x3': 'PyQt6.QtGui',
 'QMatrix3x4': 'PyQt6.QtGui',
 'QMatrix4x2': 'PyQt6.QtGui',
 'QMatrix4x3': 'PyQt6.QtGui',
 'QMatrix4x4': 'PyQt6.QtGui',
 'QMdiArea': 'PyQt6.QtWidgets',
 'QMdiSubWindow': 'PyQt6.QtWidgets',
 'QMediaCaptureSession': 'PyQt6.QtMultimedia',
 'QMediaDevices': 'PyQt6.QtMultimedia',
 'QMediaFormat': 'PyQt6.QtMultimedia',
 'QMediaMetaData': 'PyQt6.QtMultimedia',
 'QMediaPlayer': 'PyQt6.QtMultimedia',
 'QMediaRecorder': 'PyQt6.QtMultimedia',
 'QMediaTimeRange': 'PyQt6.QtMultimedia',
 'QMenu': 'PyQt6.QtWidgets',
 'QMenuBar': 'PyQt6.QtWidgets',
 'QMessageAuthenticationCode': 'PyQt6.QtCore',
 'QMessageBox': 'PyQt6.QtWidgets',
 'QMessageLogContext': 'PyQt6.QtCore',
 'QMessageLogger': 'PyQt6.QtCore',
 'QMetaClassInfo': 'PyQt6.QtCore',
 'QMetaEnum': 'PyQt6.QtCore',
 'QMetaMethod': 'PyQt6.QtCore',
 'QMetaObject': 'PyQt6.QtCore',
 'QMetaProperty': 'PyQt6.QtCore',
 'QMetaType': 'PyQt6.QtCore',
 'QMicrophonePermission': 'PyQt6.QtCore',
 'QMimeData': 'PyQt6.QtCore',
 'QMimeDatabase': 'PyQt6.QtCore',
 'QMimeType': 'PyQt6.QtCore',
 'QModelIndex': 'PyQt6.QtCore',
 'QModelRoleData': 'PyQt6.QtCore',
 'QModelRoleDataSpan': 'PyQt6.QtCore',
 'QMouseEvent': 'PyQt6.QtGui',
 'QMoveEvent': 'PyQt6.QtGui',
 'QMovie': 'PyQt6.QtGui',
 'QMutex': 'PyQt6.QtCore',
 'QMutexLocker': 'PyQt6.QtCore',
 'QNativeGestureEvent': 'PyQt6.QtGui',
 'QNativeInterface': 'PyQt6.QtQuick',
 'QNativeIpcKey': 'PyQt6.QtCore',
 'QNetworkAccessManager': 'PyQt6.QtNetwork',
 'QNetworkAddressEntry': 'PyQt6.QtNetwork',
 'QNetworkCacheMetaData': 'PyQt6.QtNetwork',
 'QNetworkCookie': 'PyQt6.QtNetwork',
 'QNetworkCookieJar': 'PyQt6.QtNetwork',
 'QNetworkDatagram': 'PyQt6.QtNetwork',
 'QNetworkDiskCache': 'PyQt6.QtNetwork',
 'QNetworkInformation': 'PyQt6.QtNetwork',
 'QNetworkInterface': 'PyQt6.QtNetwork',
 'QNetworkProxy': 'PyQt6.QtNetwork',
 'QNetworkProxyFactory': 'PyQt6.QtNetwork',
 'QNetworkProxyQuery': 'PyQt6.QtNetwork',
 'QNetworkReply': 'PyQt6.QtNetwork',
 'QNetworkRequest': 'PyQt6.QtNetwork',
 'QObject': 'PyQt6.QtCore',
 'QObjectCleanupHandler': 'PyQt6.QtCore',
 'QOcspCertificateStatus': 'PyQt6.QtNetwork',
 'QOcspResponse': 'PyQt6.QtNetwork',
 'QOcspRevocationReason': 'PyQt6.QtNetwork',
 'QOffscreenSurface': 'PyQt6.QtGui',
 'QOpenGLBuffer': 'PyQt6.QtOpenGL',
 'QOpenGLContext': 'PyQt6.QtGui',
 'QOpenGLContextGroup': 'PyQt6.QtGui',
 'QOpenGLDebugLogger': 'PyQt6.QtOpenGL',
 'QOpenGLDebugMessage': 'PyQt6.QtOpenGL',
 'QOpenGLFramebufferObject': 'PyQt6.QtOpenGL',
 'QOpenGLFramebufferObjectFormat': 'PyQt6.QtOpenGL',
 'QOpenGLFunctions_2_0': 'PyQt6.QtOpenGL',
 'QOpenGLFunctions_2_1': 'PyQt6.QtOpenGL',
 'QOpenGLFunctions_4_1_Core': 'PyQt6.QtOpenGL',
 'QOpenGLPaintDevice': 'PyQt6.QtOpenGL',
 'QOpenGLPixelTransferOptions': 'PyQt6.QtOpenGL',
 'QOpenGLShader': 'PyQt6.QtOpenGL',
 'QOpenGLShaderProgram': 'PyQt6.QtOpenGL',
 'QOpenGLTexture': 'PyQt6.QtOpenGL',
 'QOpenGLTextureBlitter': 'PyQt6.QtOpenGL',
 'QOpenGLTimeMonitor': 'PyQt6.QtOpenGL',
 'QOpenGLTimerQuery': 'PyQt6.QtOpenGL',
 'QOpenGLVersionFunctionsFactory': 'PyQt6.QtOpenGL',
 'QOpenGLVersionProfile': 'PyQt6.QtOpenGL',
 'QOpenGLVertexArrayObject': 'PyQt6.QtOpenGL',
 'QOpenGLWidget': 'PyQt6.QtOpenGLWidgets',
 'QOpenGLWindow': 'PyQt6.QtOpenGL',
 'QOperatingSystemVersion': 'PyQt6.QtCore',
 'QOperatingSystemVersionBase': 'PyQt6.QtCore',
 'QPageLayout': 'PyQt6.QtGui',
 'QPageRanges': 'PyQt6.QtGui',
 'QPageSetupDialog': 'PyQt6.QtPrintSupport',
 'QPageSize': 'PyQt6.QtGui',
 'QPagedPaintDevice': 'PyQt6.QtGui',
 'QPaintDevice': 'PyQt6.QtGui',
 'QPaintDeviceWindow': 'PyQt6.QtGui',
 'QPaintEngine': 'PyQt6.QtGui',
 'QPaintEngineState': 'PyQt6.QtGui',
 'QPaintEvent': 'PyQt6.QtGui',
 'QPainter': 'PyQt6.QtGui',
 'QPainterPath': 'PyQt6.QtGui',
 'QPainterPathStroker': 'PyQt6.QtGui',
 'QPalette': 'PyQt6.QtGui',
 'QPanGesture': 'PyQt6.QtWidgets',
 'QParallelAnimationGroup': 'PyQt6.QtCore',
 'QPasswordDigestor': 'PyQt6.QtNetwork',
 'QPauseAnimation': 'PyQt6.QtCore',
 'QPdfWriter': 'PyQt6.QtGui',
 'QPen': 'PyQt6.QtGui',
 'QPermission': 'PyQt6.QtCore',
 'QPersistentModelIndex': 'PyQt6.QtCore',
 'QPicture': 'PyQt6.QtGui',
 'QPinchGesture': 'PyQt6.QtWidgets',
 'QPixelFormat': 'PyQt6.QtGui',
 'QPixmap': 'PyQt6.QtGui',
 'QPixmapCache': 'PyQt6.QtGui',
 'QPlainTextDocumentLayout': 'PyQt6.QtWidgets',
 'QPlainTextEdit': 'PyQt6.QtWidgets',
 'QPlatformSurfaceEvent': 'PyQt6.QtGui',
 'QPluginLoader': 'PyQt6.QtCore',
 'QPoint': 'PyQt6.QtCore',
 'QPointF': 'PyQt6.QtCore',
 'QPointerEvent': 'PyQt6.QtGui',
 'QPointingDevice': 'PyQt6.QtGui',
 'QPointingDeviceUniqueId': 'PyQt6.QtGui',
 'QPolygon': 'PyQt6.QtGui',
 'QPolygonF': 'PyQt6.QtGui',
 'QPrintDialog': 'PyQt6.QtPrintSupport',
 'QPrintEngine': 'PyQt6.QtPrintSupport',
 'QPrintPreviewDialog': 'PyQt6.QtPrintSupport',
 'QPrintPreviewWidget': 'PyQt6.QtPrintSupport',
 'QPrinter': 'PyQt6.QtPrintSupport',
 'QPrinterInfo': 'PyQt6.QtPrintSupport',
 'QProcess': 'PyQt6.QtCore',
 'QProcessEnvironment': 'PyQt6.QtCore',
 'QProgressBar': 'PyQt6.QtWidgets',
 'QProgressDialog': 'PyQt6.QtWidgets',
 'QPropertyAnimation': 'PyQt6.QtCore',
 'QProxyStyle': 'PyQt6.QtWidgets',
 'QPushButton': 'PyQt6.QtWidgets',
 'QQuaternion': 'PyQt6.QtGui',
 'QQuickAsyncImageProvider': 'PyQt6.QtQuick',
 'QQuickCloseEvent': 'PyQt6.QtQuick',
 'QQuickFramebufferObject': 'PyQt6.QtQuick',
 'QQuickGraphicsConfiguration': 'PyQt6.QtQuick',
 'QQuickGraphicsDevice': 'PyQt6.QtQuick',
 'QQuickImageProvider': 'PyQt6.QtQuick',
 'QQuickImageResponse': 'PyQt6.QtQuick',
 'QQuickItem': 'PyQt6.QtQuick',
 'QQuickItemGrabResult': 'PyQt6.QtQuick',
 'QQuickPaintedItem': 'PyQt6.QtQuick',
 'QQuickRenderControl': 'PyQt6.QtQuick',
 'QQuickRenderTarget': 'PyQt6.QtQuick',
 'QQuickTextDocument': 'PyQt6.QtQuick',
 'QQuickTextureFactory': 'PyQt6.QtQuick',
 'QQuickView': 'PyQt6.QtQuick',
 'QQuickWindow': 'PyQt6.QtQuick',
 'QRadialGradient': 'PyQt6.QtGui',
 'QRadioButton': 'PyQt6.QtWidgets',
 'QRandomGenerator': 'PyQt6.QtCore',
 'QRasterWindow': 'PyQt6.QtGui',
 'QRawFont': 'PyQt6.QtGui',
 'QReadLocker': 'PyQt6.QtCore',
 'QReadWriteLock': 'PyQt6.QtCore',
 'QRect': 'PyQt6.QtCore',
 'QRectF': 'PyQt6.QtCore',
 'QRecursiveMutex': 'PyQt6.QtCore',
 'QRegion': 'PyQt6.QtGui',
 'QRegularExpression': 'PyQt6.QtCore',
 'QRegularExpressionMatch': 'PyQt6.QtCore',
 'QRegularExpressionMatchIterator': 'PyQt6.QtCore',
 'QRegularExpressionValidator': 'PyQt6.QtGui',
 'QResizeEvent': 'PyQt6.QtGui',
 'QResource': 'PyQt6.QtCore',
 'QRgba64': 'PyQt6.QtGui',
 'QRubberBand': 'PyQt6.QtWidgets',
 'QRunnable': 'PyQt6.QtCore',
 'QSGBasicGeometryNode': 'PyQt6.QtQuick',
 'QSGClipNode': 'PyQt6.QtQuick',
 'QSGDynamicTexture': 'PyQt6.QtQuick',
 'QSGFlatColorMaterial': 'PyQt6.QtQuick',
 'QSGGeometry': 'PyQt6.QtQuick',
 'QSGGeometryNode': 'PyQt6.QtQuick',
 'QSGImageNode': 'PyQt6.QtQuick',
 'QSGMaterial': 'PyQt6.QtQuick',
 'QSGMaterialShader': 'PyQt6.QtQuick',
 'QSGMaterialType': 'PyQt6.QtQuick',
 'QSGNode': 'PyQt6.QtQuick',
 'QSGOpacityNode': 'PyQt6.QtQuick',
 'QSGOpaqueTextureMaterial': 'PyQt6.QtQuick',
 'QSGRectangleNode': 'PyQt6.QtQuick',
 'QSGRenderNode': 'PyQt6.QtQuick',
 'QSGRendererInterface': 'PyQt6.QtQuick',
 'QSGSimpleRectNode': 'PyQt6.QtQuick',
 'QSGSimpleTextureNode': 'PyQt6.QtQuick',
 'QSGTextNode': 'PyQt6.QtQuick',
 'QSGTexture': 'PyQt6.QtQuick',
 'QSGTextureMaterial': 'PyQt6.QtQuick',
 'QSGTextureProvider': 'PyQt6.QtQuick',
 'QSGTransformNode': 'PyQt6.QtQuick',
 'QSGVertexColorMaterial': 'PyQt6.QtQuick',
 'QSaveFile': 'PyQt6.QtCore',
 'QScreen': 'PyQt6.QtGui',
 'QScreenCapture': 'PyQt6.QtMultimedia',
 'QScrollArea': 'PyQt6.QtWidgets',
 'QScrollBar': 'PyQt6.QtWidgets',
 'QScrollEvent': 'PyQt6.QtGui',
 'QScrollPrepareEvent': 'PyQt6.QtGui',
 'QScroller': 'PyQt6.QtWidgets',
 'QScrollerProperties': 'PyQt6.QtWidgets',
 'QSemaphore': 'PyQt6.QtCore',
 'QSemaphoreReleaser': 'PyQt6.QtCore',
 'QSequentialAnimationGroup': 'PyQt6.QtCore',
 'QSessionManager': 'PyQt6.QtGui',
 'QSettings': 'PyQt6.QtCore',
 'QSharedMemory': 'PyQt6.QtCore',
 'QShortcut': 'PyQt6.QtGui',
 'QShortcutEvent': 'PyQt6.QtGui',
 'QShowEvent': 'PyQt6.QtGui',
 'QSignalBlocker': 'PyQt6.QtCore',
 'QSignalMapper': 'PyQt6.QtCore',
 'QSinglePointEvent': 'PyQt6.QtGui',
 'QSize': 'PyQt6.QtCore',
 'QSizeF': 'PyQt6.QtCore',
 'QSizeGrip': 'PyQt6.QtWidgets',
 'QSizePolicy': 'PyQt6.QtWidgets',
 'QSlider': 'PyQt6.QtWidgets',
 'QSocketNotifier': 'PyQt6.QtCore',
 'QSortFilterProxyModel': 'PyQt6.QtCore',
 'QSoundEffect': 'PyQt6.QtMultimedia',
 'QSpacerItem': 'PyQt6.QtWidgets',
 'QSpinBox': 'PyQt6.QtWidgets',
 'QSplashScreen': 'PyQt6.QtWidgets',
 'QSplitter': 'PyQt6.QtWidgets',
 'QSplitterHandle': 'PyQt6.QtWidgets',
 'QSsl': 'PyQt6.QtNetwork',
 'QSslCertificate': 'PyQt6.QtNetwork',
 'QSslCertificateExtension': 'PyQt6.QtNetwork',
 'QSslCipher': 'PyQt6.QtNetwork',
 'QSslConfiguration': 'PyQt6.QtNetwork',
 'QSslDiffieHellmanParameters': 'PyQt6.QtNetwork',
 'QSslEllipticCurve': 'PyQt6.QtNetwork',
 'QSslError': 'PyQt6.QtNetwork',
 'QSslKey': 'PyQt6.QtNetwork',
 'QSslPreSharedKeyAuthenticator': 'PyQt6.QtNetwork',
 'QSslServer': 'PyQt6.QtNetwork',
 'QSslSocket': 'PyQt6.QtNetwork',
 'QStackedLayout': 'PyQt6.QtWidgets',
 'QStackedWidget': 'PyQt6.QtWidgets',
 'QStandardItem': 'PyQt6.QtGui',
 'QStandardItemModel': 'PyQt6.QtGui',
 'QStandardPaths': 'PyQt6.QtCore',
 'QStaticText': 'PyQt6.QtGui',
 'QStatusBar': 'PyQt6.QtWidgets',
 'QStatusTipEvent': 'PyQt6.QtGui',
 'QStorageInfo': 'PyQt6.QtCore',
 'QStringConverter': 'PyQt6.QtCore',
 'QStringConverterBase': 'PyQt6.QtCore',
 'QStringDecoder': 'PyQt6.QtCore',
 'QStringEncoder': 'PyQt6.QtCore',
 'QStringListModel': 'PyQt6.QtCore',
 'QStyle': 'PyQt6.QtWidgets',
 'QStyleFactory': 'PyQt6.QtWidgets',
 'QStyleHintReturn': 'PyQt6.QtWidgets',
 'QStyleHintReturnMask': 'PyQt6.QtWidgets',
 'QStyleHintReturnVariant': 'PyQt6.QtWidgets',
 'QStyleHints': 'PyQt6.QtGui',
 'QStyleOption': 'PyQt6.QtWidgets',
 'QStyleOptionButton': 'PyQt6.QtWidgets',
 'QStyleOptionComboBox': 'PyQt6.QtWidgets',
 'QStyleOptionComplex': 'PyQt6.QtWidgets',
 'QStyleOptionDockWidget': 'PyQt6.QtWidgets',
 'QStyleOptionFocusRect': 'PyQt6.QtWidgets',
 'QStyleOptionFrame': 'PyQt6.QtWidgets',
 'QStyleOptionGraphicsItem': 'PyQt6.QtWidgets',
 'QStyleOptionGroupBox': 'PyQt6.QtWidgets',
 'QStyleOptionHeader': 'PyQt6.QtWidgets',
 'QStyleOptionHeaderV2': 'PyQt6.QtWidgets',
 'QStyleOptionMenuItem': 'PyQt6.QtWidgets',
 'QStyleOptionProgressBar': 'PyQt6.QtWidgets',
 'QStyleOptionRubberBand': 'PyQt6.QtWidgets',
 'QStyleOptionSizeGrip': 'PyQt6.QtWidgets',
 'QStyleOptionSlider': 'PyQt6.QtWidgets',
 'QStyleOptionSpinBox': 'PyQt6.QtWidgets',
 'QStyleOptionTab': 'PyQt6.QtWidgets',
 'QStyleOptionTabBarBase': 'PyQt6.QtWidgets',
 'QStyleOptionTabWidgetFrame': 'PyQt6.QtWidgets',
 'QStyleOptionTitleBar': 'PyQt6.QtWidgets',
 'QStyleOptionToolBar': 'PyQt6.QtWidgets',
 'QStyleOptionToolBox': 'PyQt6.QtWidgets',
 'QStyleOptionToolButton': 'PyQt6.QtWidgets',
 'QStyleOptionViewItem': 'PyQt6.QtWidgets',
 'QStylePainter': 'PyQt6.QtWidgets',
 'QStyledItemDelegate': 'PyQt6.QtWidgets',
 'QSurface': 'PyQt6.QtGui',
 'QSurfaceFormat': 'PyQt6.QtGui',
 'QSvgGenerator': 'PyQt6.QtSvg',
 'QSvgRenderer': 'PyQt6.QtSvg',
 'QSwipeGesture': 'PyQt6.QtWidgets',
 'QSyntaxHighlighter': 'PyQt6.QtGui',
 'QSysInfo': 'PyQt6.QtCore',
 'QSystemSemaphore': 'PyQt6.QtCore',
 'QSystemTrayIcon': 'PyQt6.QtWidgets',
 'QT_TRANSLATE_NOOP': 'PyQt6.QtCore',
 'QT_TR_NOOP': 'PyQt6.QtCore',
 'QT_VERSION': 'PyQt6.QtCore',
 'QT_VERSION_STR': 'PyQt6.QtCore',
 'QTabBar': 'PyQt6.QtWidgets',
 'QTabWidget': 'PyQt6.QtWidgets',
 'QTableView': 'PyQt6.QtWidgets',
 'QTableWidget': 'PyQt6.QtWidgets',
 'QTableWidgetItem': 'PyQt6.QtWidgets',
 'QTableWidgetSelectionRange': 'PyQt6.QtWidgets',
 'QTabletEvent': 'PyQt6.QtGui',
 'QTapAndHoldGesture': 'PyQt6.QtWidgets',
 'QTapGesture': 'PyQt6.QtWidgets',
 'QTcpServer': 'PyQt6.QtNetwork',
 'QTcpSocket': 'PyQt6.QtNetwork',
 'QTemporaryDir': 'PyQt6.QtCore',
 'QTemporaryFile': 'PyQt6.QtCore',
 'QTextBlock': 'PyQt6.QtGui',
 'QTextBlockFormat': 'PyQt6.QtGui',
 'QTextBlockGroup': 'PyQt6.QtGui',
 'QTextBlockUserData': 'PyQt6.QtGui',
 'QTextBoundaryFinder': 'PyQt6.QtCore',
 'QTextBrowser': 'PyQt6.QtWidgets',
 'QTextCharFormat': 'PyQt6.QtGui',
 'QTextCursor': 'PyQt6.QtGui',
 'QTextDocument': 'PyQt6.QtGui',
 'QTextDocumentFragment': 'PyQt6.QtGui',
 'QTextDocumentWriter': 'PyQt6.QtGui',
 'QTextEdit': 'PyQt6.QtWidgets',
 'QTextFormat': 'PyQt6.QtGui',
 'QTextFragment': 'PyQt6.QtGui',
 'QTextFrame': 'PyQt6.QtGui',
 'QTextFrameFormat': 'PyQt6.QtGui',
 'QTextImageFormat': 'PyQt6.QtGui',
 'QTextInlineObject': 'PyQt6.QtGui',
 'QTextItem': 'PyQt6.QtGui',
 'QTextLayout': 'PyQt6.QtGui',
 'QTextLength': 'PyQt6.QtGui',
 'QTextLine': 'PyQt6.QtGui',
 'QTextList': 'PyQt6.QtGui',
 'QTextListFormat': 'PyQt6.QtGui',
 'QTextObject': 'PyQt6.QtGui',
 'QTextObjectInterface': 'PyQt6.QtGui',
 'QTextOption': 'PyQt6.QtGui',
 'QTextStream': 'PyQt6.QtCore',
 'QTextStreamManipulator': 'PyQt6.QtCore',
 'QTextTable': 'PyQt6.QtGui',
 'QTextTableCell': 'PyQt6.QtGui',
 'QTextTableCellFormat': 'PyQt6.QtGui',
 'QTextTableFormat': 'PyQt6.QtGui',
 'QTextToSpeech': 'PyQt6.QtTextToSpeech',
 'QThread': 'PyQt6.QtCore',
 'QThreadPool': 'PyQt6.QtCore',
 'QTime': 'PyQt6.QtCore',
 'QTimeEdit': 'PyQt6.QtWidgets',
 'QTimeLine': 'PyQt6.QtCore',
 'QTimeZone': 'PyQt6.QtCore',
 'QTimer': 'PyQt6.QtCore',
 'QTimerEvent': 'PyQt6.QtCore',
 'QToolBar': 'PyQt6.QtWidgets',
 'QToolBox': 'PyQt6.QtWidgets',
 'QToolButton': 'PyQt6.QtWidgets',
 'QToolTip': 'PyQt6.QtWidgets',
 'QTouchEvent': 'PyQt6.QtGui',
 'QTransform': 'PyQt6.QtGui',
 'QTranslator': 'PyQt6.QtCore',
 'QTransposeProxyModel': 'PyQt6.QtCore',
 'QTreeView': 'PyQt6.QtWidgets',
 'QTreeWidget': 'PyQt6.QtWidgets',
 'QTreeWidgetItem': 'PyQt6.QtWidgets',
 'QTreeWidgetItemIterator': 'PyQt6.QtWidgets',
 'QTypeRevision': 'PyQt6.QtCore',
 'QUdpSocket': 'PyQt6.QtNetwork',
 'QUndoCommand': 'PyQt6.QtGui',
 'QUndoGroup': 'PyQt6.QtGui',
 'QUndoStack': 'PyQt6.QtGui',
 'QUndoView': 'PyQt6.QtWidgets',
 'QUrl': 'PyQt6.QtCore',
 'QUrlQuery': 'PyQt6.QtCore',
 'QUuid': 'PyQt6.QtCore',
 'QVBoxLayout': 'PyQt6.QtWidgets',
 'QValidator': 'PyQt6.QtGui',
 'QVariant': 'PyQt6.QtCore',
 'QVariantAnimation': 'PyQt6.QtCore',
 'QVector2D': 'PyQt6.QtGui',
 'QVector3D': 'PyQt6.QtGui',
 'QVector4D': 'PyQt6.QtGui',
 'QVersionNumber': 'PyQt6.QtCore',
 'QVideoFrame': 'PyQt6.QtMultimedia',
 'QVideoFrameFormat': 'PyQt6.QtMultimedia',
 'QVideoSink': 'PyQt6.QtMultimedia',
 'QVideoWidget': 'PyQt6.QtMultimediaWidgets',
 'QVoice': 'PyQt6.QtTextToSpeech',
 'QWIDGETSIZE_MAX': 'PyQt6.QtWidgets',
 'QWaitCondition': 'PyQt6.QtCore',
 'QWhatsThis': 'PyQt6.QtWidgets',
 'QWhatsThisClickedEvent': 'PyQt6.QtGui',
 'QWheelEvent': 'PyQt6.QtGui',
 'QWidget': 'PyQt6.QtWidgets',
 'QWidgetAction': 'PyQt6.QtWidgets',
 'QWidgetItem': 'PyQt6.QtWidgets',
 'QWindow': 'PyQt6.QtGui',
 'QWindowCapture': 'PyQt6.QtMultimedia',
 'QWindowStateChangeEvent': 'PyQt6.QtGui',
 'QWizard': 'PyQt6.QtWidgets',
 'QWizardPage': 'PyQt6.QtWidgets',
 'QWriteLocker': 'PyQt6.QtCore',
 'QXmlStreamAttribute': 'PyQt6.QtCore',
 'QXmlStreamAttributes': 'PyQt6.QtCore',
 'QXmlStreamEntityDeclaration': 'PyQt6.QtCore',
 'QXmlStreamEntityResolver': 'PyQt6.QtCore',
 'QXmlStreamNamespaceDeclaration': 'PyQt6.QtCore',
 'QXmlStreamNotationDeclaration': 'PyQt6.QtCore',
 'QXmlStreamReader': 'PyQt6.QtCore',
 'QXmlStreamWriter': 'PyQt6.QtCore',
 'Q_ARG': 'PyQt6.QtCore',
 'Q_RETURN_ARG': 'PyQt6.QtCore',
 'Qt': 'PyQt6.QtCore',
 'QtCore': 'PyQt6.QtCore',
 'QtGui': 'PyQt6.QtGui',
 'QtMsgType': 'PyQt6.QtCore',
 'QtMultimedia': 'PyQt6.QtMultimedia',
 'QtMultimediaWidgets': 'PyQt6.QtMultimediaWidgets',
 'QtNetwork': 'PyQt6.QtNetwork',
 'QtOpenGL': 'PyQt6.QtOpenGL',
 'QtOpenGLWidgets': 'PyQt6.QtOpenGLWidgets',
 'QtPrintSupport': 'PyQt6.QtPrintSupport',
 'QtQuick': 'PyQt6.QtQuick',
 'QtSvg': 'PyQt6.QtSvg',
 'QtTextToSpeech': 'PyQt6.QtTextToSpeech',
 'QtVideo': 'PyQt6.QtMultimedia',
 'QtWidgets': 'PyQt6.QtWidgets',
 'pyqtBoundSignal': 'PyQt6.QtCore',
 'pyqtClassInfo': 'PyQt6.QtCore',
 'pyqtEnum': 'PyQt6.QtCore',
 'pyqtPickleProtocol': 'PyQt6.QtCore',
 'pyqtProperty': 'PyQt6.QtCore',
 'pyqtRemoveInputHook': 'PyQt6.QtCore',
 'pyqtRestoreInputHook': 'PyQt6.QtCore',
 'pyqtSetPickleProtocol': 'PyQt6.QtCore',
 'pyqtSignal': 'PyQt6.QtCore',
 'pyqtSlot': 'PyQt6.QtCore',
 'qAbs': 'PyQt6.QtCore',
 'qAddPostRoutine': 'PyQt6.QtCore',
 'qAddPreRoutine': 'PyQt6.QtCore',
 'qAlpha': 'PyQt6.QtGui',
 'qBlue': 'PyQt6.QtGui',
 'qChecksum': 'PyQt6.QtCore',
 'qCompress': 'PyQt6.QtCore',
 'qCritical': 'PyQt6.QtCore',
 'qDebug': 'PyQt6.QtCore',
 'qDrawBorderPixmap': 'PyQt6.QtWidgets',
 'qDrawPlainRect': 'PyQt6.QtWidgets',
 'qDrawPlainRoundedRect': 'PyQt6.QtWidgets',
 'qDrawShadeLine': 'PyQt6.QtWidgets',
 'qDrawShadePanel': 'PyQt6.QtWidgets',
 'qDrawShadeRect': 'PyQt6.QtWidgets',
 'qDrawWinButton': 'PyQt6.QtWidgets',
 'qDrawWinPanel': 'PyQt6.QtWidgets',
 'qEnvironmentVariable': 'PyQt6.QtCore',
 'qEnvironmentVariableIntValue': 'PyQt6.QtCore',
 'qEnvironmentVariableIsEmpty': 'PyQt6.QtCore',
 'qEnvironmentVariableIsSet': 'PyQt6.QtCore',
 'qFatal': 'PyQt6.QtCore',
 'qFloatDistance': 'PyQt6.QtCore',
 'qFormatLogMessage': 'PyQt6.QtCore',
 'qFuzzyCompare': 'PyQt6.QtCore',
 'qFuzzyIsNull': 'PyQt6.QtCore',
 'qGray': 'PyQt6.QtGui',
 'qGreen': 'PyQt6.QtGui',
 'qInf': 'PyQt6.QtCore',
 'qInfo': 'PyQt6.QtCore',
 'qInstallMessageHandler': 'PyQt6.QtCore',
 'qIsFinite': 'PyQt6.QtCore',
 'qIsInf': 'PyQt6.QtCore',
 'qIsNaN': 'PyQt6.QtCore',
 'qPixelFormatAlpha': 'PyQt6.QtGui',
 'qPixelFormatCmyk': 'PyQt6.QtGui',
 'qPixelFormatGrayscale': 'PyQt6.QtGui',
 'qPixelFormatHsl': 'PyQt6.QtGui',
 'qPixelFormatHsv': 'PyQt6.QtGui',
 'qPixelFormatRgba': 'PyQt6.QtGui',
 'qPixelFormatYuv': 'PyQt6.QtGui',
 'qPremultiply': 'PyQt6.QtGui',
 'qQNaN': 'PyQt6.QtCore',
 'qRed': 'PyQt6.QtGui',
 'qRegisterResourceData': 'PyQt6.QtCore',
 'qRemovePostRoutine': 'PyQt6.QtCore',
 'qRgb': 'PyQt6.QtGui',
 'qRgba': 'PyQt6.QtGui',
 'qRgba64': 'PyQt6.QtGui',
 'qRound': 'PyQt6.QtCore',
 'qRound64': 'PyQt6.QtCore',
 'qSNaN': 'PyQt6.QtCore',
 'qSetFieldWidth': 'PyQt6.QtCore',
 'qSetMessagePattern': 'PyQt6.QtCore',
 'qSetPadChar': 'PyQt6.QtCore',
 'qSetRealNumberPrecision': 'PyQt6.QtCore',
 'qUncompress': 'PyQt6.QtCore',
 'qUnpremultiply': 'PyQt6.QtGui',
 'qUnregisterResourceData': 'PyQt6.QtCore',
 'qVersion': 'PyQt6.QtCore',
 'qWarning': 'PyQt6.QtCore',
 'qYieldCpu': 'PyQt6.QtCore',
 'qt_set_sequence_auto_mnemonic': 'PyQt6.QtGui',
 'sip': 'PyQt6.sip'}
module_names = frozenset(('QtCore',
 'QtGui',
 'QtWidgets',
 'QtNetwork',
 'QtSvg',
 'QtPrintSupport',
 'QtOpenGL',
 'QtOpenGLWidgets',
 'QtQuick',
 'QtMultimedia',
 'QtMultimediaWidgets',
 'QtTextToSpeech',
 'sip')
)



#!/usr/local/bin/python3.13


__license__   = 'GPL v3'
__copyright__ = '2012, Kovid Goyal <kovid@kovidgoyal.net>'
__docformat__ = 'restructuredtext en'

import io
import os
import re
import shutil
import struct
import textwrap

from lxml import etree, html

from calibre import guess_type, replace_entities, xml_replace_entities
from calibre.ebooks import DRMError, unit_convert
from calibre.ebooks.chardet import strip_encoding_declarations
from calibre.ebooks.compression.palmdoc import decompress_doc
from calibre.ebooks.metadata import MetaInformation
from calibre.ebooks.metadata.opf2 import OPF, OPFCreator
from calibre.ebooks.metadata.toc import TOC
from calibre.ebooks.mobi import MobiError
from calibre.ebooks.mobi.huffcdic import HuffReader
from calibre.ebooks.mobi.reader.headers import BookHeader
from calibre.utils.cleantext import clean_ascii_chars, clean_xml_chars
from calibre.utils.img import AnimatedGIF, gif_data_to_png_data, save_cover_data_to
from calibre.utils.imghdr import what
from calibre.utils.logging import default_log
from polyglot.builtins import iteritems


class TopazError(ValueError):
    pass


class KFXError(ValueError):

    def __init__(self):
        ValueError.__init__(self, _(
            'This is an Amazon KFX book. It cannot be processed.'
            ' See {} for information on how to handle KFX books.'
        ).format('https://www.mobileread.com/forums/showthread.php?t=283371'))


class MobiReader:
    PAGE_BREAK_PAT = re.compile(
        r'<\s*/{0,1}\s*mbp:pagebreak((?:\s+[^/>]*){0,1})/{0,1}\s*>\s*(?:<\s*/{0,1}\s*mbp:pagebreak\s*/{0,1}\s*>)*',
        re.IGNORECASE)
    IMAGE_ATTRS = ('lowrecindex', 'recindex', 'hirecindex')

    def __init__(self, filename_or_stream, log=None, user_encoding=None, debug=None,
            try_extra_data_fix=False):
        self.log = log or default_log
        self.debug = debug
        self.embedded_mi = None
        self.warned_about_trailing_entry_corruption = False
        self.base_css_rules = textwrap.dedent('''
                body { text-align: justify }

                blockquote { margin: 0em 0em 0em 2em; }

                p { margin: 0em; text-indent: 1.5em }

                .bold { font-weight: bold }

                .italic { font-style: italic }

                .underline { text-decoration: underline }

                .mbp_pagebreak {
                    page-break-after: always; margin: 0; display: block
                }
                ''')
        self.tag_css_rules = {}
        self.left_margins = {}
        self.text_indents = {}

        if hasattr(filename_or_stream, 'read'):
            stream = filename_or_stream
            stream.seek(0)
        else:
            stream = open(filename_or_stream, 'rb')

        raw = stream.read()
        if raw.startswith(b'TPZ'):
            raise TopazError(_('This is an Amazon Topaz book. It cannot be processed.'))
        if raw.startswith(b'\xeaDRMION\xee'):
            raise KFXError()

        self.header   = raw[0:72]
        self.name     = self.header[:32].replace(b'\x00', b'')
        self.num_sections, = struct.unpack('>H', raw[76:78])

        self.ident = self.header[0x3C:0x3C + 8].upper()
        if self.ident not in (b'BOOKMOBI', b'TEXTREAD'):
            raise MobiError(f'Unknown book type: {self.ident!r}')

        self.sections = []
        self.section_headers = []
        for i in range(self.num_sections):
            offset, a1, a2, a3, a4 = struct.unpack('>LBBBB', raw[78 + i * 8:78 + i * 8 + 8])
            flags, val = a1, a2 << 16 | a3 << 8 | a4
            self.section_headers.append((offset, flags, val))

        def section(section_number):
            if section_number == self.num_sections - 1:
                end_off = len(raw)
            else:
                end_off = self.section_headers[section_number + 1][0]
            off = self.section_headers[section_number][0]
            return raw[off:end_off]

        for i in range(self.num_sections):
            self.sections.append((section(i), self.section_headers[i]))

        self.book_header = bh = BookHeader(self.sections[0][0], self.ident,
            user_encoding, self.log, try_extra_data_fix=try_extra_data_fix)
        self.name = self.name.decode(self.book_header.codec, 'replace')
        self.kf8_type = None
        k8i = getattr(self.book_header.exth, 'kf8_header', None)

        # Ancient PRC files from Baen can have random values for
        # mobi_version, so be conservative
        if (self.book_header.mobi_version == 8 and hasattr(self.book_header,
            'skelidx')):
            self.kf8_type = 'standalone'
        elif k8i is not None:  # Check for joint mobi 6 and kf 8 file
            try:
                raw = self.sections[k8i-1][0]
            except:
                raw = None
            if raw == b'BOUNDARY':
                try:
                    self.book_header = BookHeader(self.sections[k8i][0],
                            self.ident, user_encoding, self.log)
                    self.book_header.kf8_first_image_index = self.book_header.first_image_index + k8i
                    self.book_header.mobi6_records = bh.records

                    # Need the first_image_index from the mobi 6 header as well
                    for x in ('first_image_index',):
                        setattr(self.book_header, x, getattr(bh, x))

                    # We need to do this because the MOBI 6 text extract code
                    # does not know anything about the kf8 offset
                    if hasattr(self.book_header, 'huff_offset'):
                        self.book_header.huff_offset += k8i

                    self.kf8_type = 'joint'
                    self.kf8_boundary = k8i-1
                except:
                    self.book_header = bh

    def check_for_drm(self):
        if self.book_header.encryption_type != 0:
            try:
                name = self.book_header.exth.mi.title
            except:
                name = self.name
            if not name:
                name = self.name
            raise DRMError(name)

    def extract_content(self, output_dir, parse_cache):
        output_dir = os.path.abspath(output_dir)
        self.check_for_drm()
        processed_records = self.extract_text()
        if self.debug is not None:
            parse_cache['calibre_raw_mobi_markup'] = self.mobi_html
        self.add_anchors()
        self.processed_html = self.processed_html.decode(self.book_header.codec,
            'ignore')
        self.processed_html = self.processed_html.replace('</</', '</')
        self.processed_html = re.sub(r'</([a-zA-Z]+)<', r'</\1><',
                self.processed_html)
        self.processed_html = self.processed_html.replace('\ufeff', '')
        # Remove tags of the form <xyz: ...> as they can cause issues further
        # along the pipeline
        self.processed_html = re.sub(r'</{0,1}[a-zA-Z]+:\s+[^>]*>', '',
                self.processed_html)

        self.processed_html = strip_encoding_declarations(self.processed_html)
        self.processed_html = xml_replace_entities(self.processed_html)
        image_name_map = self.extract_images(processed_records, output_dir)
        self.replace_page_breaks()
        self.cleanup_html()

        self.log.debug('Parsing HTML...')
        self.processed_html = clean_xml_chars(self.processed_html)
        try:
            root = html.fromstring(self.processed_html)
            if len(root.xpath('//html')) > 5:
                root = html.fromstring(self.processed_html.replace('\x0c',
                    '').replace('\x14', ''))
        except Exception:
            self.log.warning('MOBI markup appears to contain random bytes. Stripping.')
            self.processed_html = self.remove_random_bytes(self.processed_html)
            try:
                root = html.fromstring(self.processed_html)
            except Exception:
                self.log.warning('MOBI markup could not be parsed by lxml using html5-parser')
                # Happens on windows with python 3 where lxml causes libxml to die with an
                # error about using UCS-4 little endian encoding if certain
                # characters are present in the input
                from html5_parser import parse
                root = parse(self.processed_html, keep_doctype=False, namespace_elements=False, maybe_xhtml=False, sanitize_names=True)
        if root.xpath('descendant::p/descendant::p'):
            from html5_parser import parse
            self.log.warning('Malformed markup, parsing using html5-parser')
            self.processed_html = strip_encoding_declarations(self.processed_html)
            # These trip up the html5 parser causing all content to be placed
            # under the <guide> tag
            self.processed_html = re.sub(r'<metadata>.+?</metadata>', '', self.processed_html, flags=re.I)
            self.processed_html = re.sub(r'<guide>.+?</guide>', '', self.processed_html, flags=re.I)
            try:
                root = parse(self.processed_html, maybe_xhtml=False, keep_doctype=False, sanitize_names=True)
            except Exception:
                self.log.warning('MOBI markup appears to contain random bytes. Stripping.')
                self.processed_html = self.remove_random_bytes(self.processed_html)
                root = parse(self.processed_html, maybe_xhtml=False, keep_doctype=False, sanitize_names=True)
            if len(root.xpath('body/descendant::*')) < 1:
                # There are probably stray </html>s in the markup
                self.processed_html = self.processed_html.replace('</html>',
                        '')
                root = parse(self.processed_html, maybe_xhtml=False, keep_doctype=False, sanitize_names=True)

        if root.tag != 'html':
            self.log.warn('File does not have opening <html> tag')
            nroot = html.fromstring('<html><head></head><body></body></html>')
            bod = nroot.find('body')
            for child in list(root):
                child.getparent().remove(child)
                bod.append(child)
            root = nroot

        htmls = list(root.xpath('//html'))

        if len(htmls) > 1:
            self.log.warn('Markup contains multiple <html> tags, merging.')
            # Merge all <head> and <body> sections
            for h in htmls:
                p = h.getparent()
                if hasattr(p, 'remove'):
                    p.remove(h)
            bodies, heads = root.xpath('//body'), root.xpath('//head')
            for x in root:
                root.remove(x)
            head, body = map(root.makeelement, ('head', 'body'))
            for h in heads:
                for x in h:
                    h.remove(x)
                    head.append(x)
            for b in bodies:
                for x in b:
                    b.remove(x)
                    body.append(x)
            root.append(head), root.append(body)
        for x in root.xpath('//script'):
            x.getparent().remove(x)

        head = root.xpath('//head')
        if head:
            head = head[0]
        else:
            head = root.makeelement('head', {})
            root.insert(0, head)
        head.text = '\n\t'
        link = head.makeelement('link', {'type':'text/css',
            'href':'styles.css', 'rel':'stylesheet'})
        head.insert(0, link)
        link.tail = '\n\t'
        title = head.xpath('descendant::title')
        m = head.makeelement('meta', {'http-equiv':'Content-Type',
            'content':'text/html; charset=utf-8'})
        head.insert(0, m)
        if not title:
            title = head.makeelement('title', {})
            try:
                title.text = self.book_header.title
            except ValueError:
                title.text = clean_ascii_chars(self.book_header.title)
            title.tail = '\n\t'
            head.insert(0, title)
            head.text = '\n\t'

        self.upshift_markup(root, image_name_map)
        guides = root.xpath('//guide')
        guide = guides[0] if guides else None
        metadata_elems = root.xpath('//metadata')
        if metadata_elems and self.book_header.exth is None:
            self.read_embedded_metadata(root, metadata_elems[0], guide)
        for elem in guides + metadata_elems:
            elem.getparent().remove(elem)
        htmlfile = os.path.join(output_dir, 'index.html')
        try:
            for ref in guide.xpath('descendant::reference'):
                if 'href' in ref.attrib:
                    ref.attrib['href'] = os.path.basename(htmlfile) + ref.attrib['href']
        except AttributeError:
            pass

        def write_as_utf8(path, data):
            if isinstance(data, str):
                data = data.encode('utf-8')
            with open(path, 'wb') as f:
                f.write(data)

        parse_cache[htmlfile] = root
        self.htmlfile = htmlfile
        ncx = io.BytesIO()
        opf, ncx_manifest_entry = self.create_opf(htmlfile, guide, root)
        self.created_opf_path = os.path.splitext(htmlfile)[0] + '.opf'
        opf.render(open(self.created_opf_path, 'wb'), ncx,
            ncx_manifest_entry=ncx_manifest_entry)
        ncx = ncx.getvalue()
        if ncx:
            ncx_path = os.path.join(os.path.dirname(htmlfile), 'toc.ncx')
            write_as_utf8(ncx_path, ncx)

        css = [self.base_css_rules, '\n\n']
        for cls, rule in self.tag_css_rules.items():
            css.append(f'.{cls} {{ {rule} }}\n\n')
        write_as_utf8('styles.css', ''.join(css))

        if self.book_header.exth is not None or self.embedded_mi is not None:
            self.log.debug('Creating OPF...')
            ncx = io.BytesIO()
            opf, ncx_manifest_entry  = self.create_opf(htmlfile, guide, root)
            opf.render(open(os.path.splitext(htmlfile)[0] + '.opf', 'wb'), ncx,
                ncx_manifest_entry)
            ncx = ncx.getvalue()
            if ncx:
                write_as_utf8(os.path.splitext(htmlfile)[0] + '.ncx', ncx)

    def read_embedded_metadata(self, root, elem, guide):
        raw = b'<?xml version="1.0" encoding="utf-8" ?>\n<package>' + \
                html.tostring(elem, encoding='utf-8') + b'</package>'
        stream = io.BytesIO(raw)
        opf = OPF(stream)
        self.embedded_mi = opf.to_book_metadata()
        if guide is not None:
            for ref in guide.xpath('descendant::reference'):
                if 'cover' in ref.get('type', '').lower():
                    href = ref.get('href', '')
                    if href.startswith('#'):
                        href = href[1:]
                    anchors = root.xpath(f'//*[@id="{href}"]')
                    if anchors:
                        cpos = anchors[0]
                        reached = False
                        for elem in root.iter():
                            if elem is cpos:
                                reached = True
                            if reached and elem.tag == 'img':
                                cover = elem.get('src', None)
                                self.embedded_mi.cover = cover
                                elem.getparent().remove(elem)
                                break
                    break

    def cleanup_html(self):
        self.log.debug('Cleaning up HTML...')
        self.processed_html = re.sub(r'<div height="0(pt|px|ex|em|%){0,1}"></div>', '', self.processed_html)
        if self.book_header.ancient and b'<html' not in self.mobi_html[:300].lower():
            self.processed_html = '<html><p>' + self.processed_html.replace('\n\n', '<p>') + '</html>'
        self.processed_html = self.processed_html.replace('\r\n', '\n')
        self.processed_html = self.processed_html.replace('> <', '>\n<')
        self.processed_html = self.processed_html.replace('<mbp: ', '<mbp:')
        self.processed_html = re.sub(r'<\?xml[^>]*>', '', self.processed_html)
        self.processed_html = re.sub(r'<\s*(/?)\s*o:p[^>]*>', '', self.processed_html)
        # Swap inline and block level elements, and order block level elements according to priority
        # - lxml and beautifulsoup expect/assume a specific order based on xhtml spec
        self.processed_html = re.sub(
            r'(?i)(?P<styletags>(<(h\d+|i|b|u|em|small|big|strong|tt)>\s*){1,})(?P<para><p[^>]*>)', r'\g<para>\g<styletags>', self.processed_html)
        self.processed_html = re.sub(
            r'(?i)(?P<para></p[^>]*>)\s*(?P<styletags>(</(h\d+|i|b|u|em|small|big|strong|tt)>\s*){1,})', r'\g<styletags>\g<para>', self.processed_html)
        self.processed_html = re.sub(
            r'(?i)(?P<blockquote>(</(blockquote|div)[^>]*>\s*){1,})(?P<para></p[^>]*>)', r'\g<para>\g<blockquote>', self.processed_html)
        self.processed_html = re.sub(
            r'(?i)(?P<para><p[^>]*>)\s*(?P<blockquote>(<(blockquote|div)[^>]*>\s*){1,})', r'\g<blockquote>\g<para>', self.processed_html)
        bods = htmls = 0
        for x in re.finditer(r'</body>|</html>', self.processed_html):
            if x == '</body>':
                bods +=1
            else:
                htmls += 1
            if bods > 1 and htmls > 1:
                break
        if bods > 1:
            self.processed_html = self.processed_html.replace('</body>', '')
        if htmls > 1:
            self.processed_html = self.processed_html.replace('</html>', '')

    def remove_random_bytes(self, html):
        return re.sub(r'\x14|\x15|\x19|\x1c|\x1d|\xef|\x12|\x13|\xec|\x08|\x01|\x02|\x03|\x04|\x05|\x06|\x07', '', html)

    def ensure_unit(self, raw, unit='px'):
        if re.search(r'\d+$', raw) is not None:
            raw += unit
        return raw

    def upshift_markup(self, root, image_name_map=None):
        self.log.debug('Converting style information to CSS...')
        image_name_map = image_name_map or {}
        size_map = {
            'xx-small': '0.5',
            'x-small': '1',
            'small': '2',
            'medium': '3',
            'large': '4',
            'x-large': '5',
            'xx-large': '6',
            }

        def barename(x):
            return x.rpartition(':')[-1]

        mobi_version = self.book_header.mobi_version
        for x in root.xpath('//ncx'):
            x.getparent().remove(x)
        svg_tags = []
        forwardable_anchors = []
        pagebreak_anchors = []
        BLOCK_TAGS = {'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'div', 'p'}
        for i, tag in enumerate(root.iter(etree.Element)):
            tag.attrib.pop('xmlns', '')
            for x in tag.attrib:
                if ':' in x:
                    del tag.attrib[x]
            if tag.tag and barename(tag.tag) == 'svg':
                svg_tags.append(tag)
            if tag.tag and barename(tag.tag.lower()) in \
                ('country-region', 'place', 'placetype', 'placename',
                    'state', 'city', 'street', 'address', 'content', 'form'):
                tag.tag = 'div' if tag.tag in ('content', 'form') else 'span'
                for key in tag.attrib.keys():
                    tag.attrib.pop(key)
                continue
            styles, attrib = [], tag.attrib
            if 'style' in attrib:
                style = attrib.pop('style').strip()
                if style:
                    styles.append(style)
            if 'height' in attrib:
                height = attrib.pop('height').strip()
                if (
                        height and '<' not in height and '>' not in height and
                        re.search(r'\d+', height)):
                    if tag.tag in ('table', 'td', 'tr'):
                        pass
                    elif tag.tag == 'img':
                        tag.set('height', height)
                    else:
                        if tag.tag == 'div' and not tag.text and \
                                (not tag.tail or not tag.tail.strip()) and \
                                not len(list(tag.iterdescendants())):
                            # Paragraph spacer
                            # Insert nbsp so that the element is never
                            # discarded by a renderer
                            tag.text = '\u00a0'  # nbsp
                            styles.append(f'height: {self.ensure_unit(height)}')
                        else:
                            styles.append(f'margin-top: {self.ensure_unit(height)}')
            if 'width' in attrib:
                width = attrib.pop('width').strip()
                if width and re.search(r'\d+', width):
                    if tag.tag in ('table', 'td', 'tr'):
                        pass
                    elif tag.tag == 'img':
                        tag.set('width', width)
                    else:
                        ewidth = self.ensure_unit(width)
                        styles.append(f'text-indent: {ewidth}')
                        try:
                            ewidth_val = unit_convert(ewidth, 12, 500, 166)
                            self.text_indents[tag] = ewidth_val
                        except:
                            pass
                        if width.startswith('-'):
                            styles.append(f'margin-left: {self.ensure_unit(width[1:])}')
                            try:
                                ewidth_val = unit_convert(ewidth[1:], 12, 500, 166)
                                self.left_margins[tag] = ewidth_val
                            except:
                                pass

            if 'align' in attrib:
                align = attrib.pop('align').strip()
                if align:
                    align = align.lower()
                    if align == 'baseline':
                        styles.append('vertical-align: '+align)
                    else:
                        styles.append(f'text-align: {align}')
            if tag.tag == 'hr':
                if mobi_version == 1:
                    tag.tag = 'div'
                    styles.append('page-break-before: always')
                    styles.append('display: block')
                    styles.append('margin: 0')
            elif tag.tag == 'i':
                tag.tag = 'span'
                tag.attrib['class'] = 'italic'
            elif tag.tag == 'u':
                tag.tag = 'span'
                tag.attrib['class'] = 'underline'
            elif tag.tag == 'b':
                tag.tag = 'span'
                tag.attrib['class'] = 'bold'
            elif tag.tag == 'font':
                sz = tag.get('size', '').lower()
                try:
                    float(sz)
                except ValueError:
                    if sz in list(size_map.keys()):
                        attrib['size'] = size_map[sz]
            elif tag.tag == 'img':
                recindex = None
                for attr in self.IMAGE_ATTRS:
                    recindex = attrib.pop(attr, None) or recindex
                if recindex is not None:
                    try:
                        recindex = int(recindex)
                    except Exception:
                        pass
                    else:
                        attrib['src'] = 'images/' + image_name_map.get(recindex, f'{recindex:05}.jpg')
                for attr in ('width', 'height'):
                    if attr in attrib:
                        val = attrib[attr]
                        if val.lower().endswith('em'):
                            try:
                                nval = float(val[:-2])
                                nval *= 16 * (168.451/72)  # Assume this was set using the Kindle profile
                                attrib[attr] = f'{int(nval)}px'
                            except:
                                del attrib[attr]
                        elif val.lower().endswith('%'):
                            del attrib[attr]
            elif tag.tag == 'pre':
                if not tag.text:
                    tag.tag = 'div'

            if (attrib.get('class', None) == 'mbp_pagebreak' and tag.tag ==
                    'div' and 'filepos-id' in attrib):
                pagebreak_anchors.append(tag)

            if 'color' in attrib:
                styles.append('color: ' + attrib.pop('color'))
            if 'bgcolor' in attrib:
                styles.append('background-color: ' + attrib.pop('bgcolor'))

            if 'filepos-id' in attrib:
                attrib['id'] = attrib.pop('filepos-id')
                if 'name' in attrib and attrib['name'] != attrib['id']:
                    attrib['name'] = attrib['id']
            if 'filepos' in attrib:
                filepos = attrib.pop('filepos')
                try:
                    attrib['href'] = f'#filepos{int(filepos)}'
                except ValueError:
                    pass
            if (tag.tag == 'a' and attrib.get('id', '').startswith('filepos') and
                    not tag.text and len(tag) == 0 and (tag.tail is None or not
                        tag.tail.strip()) and getattr(tag.getnext(), 'tag',
                            None) in BLOCK_TAGS):
                # This is an empty anchor immediately before a block tag, move
                # the id onto the block tag instead
                forwardable_anchors.append(tag)

            if styles:
                ncls = None
                rule = '; '.join(styles)
                for sel, srule in self.tag_css_rules.items():
                    if srule == rule:
                        ncls = sel
                        break
                if ncls is None:
                    ncls = f'calibre_{i}'
                    self.tag_css_rules[ncls] = rule
                cls = attrib.get('class', '')
                cls = cls + (' ' if cls else '') + ncls
                attrib['class'] = cls

        for tag in svg_tags:
            images = tag.xpath('descendant::img[@src]')
            parent = tag.getparent()

            if images and hasattr(parent, 'find'):
                index = parent.index(tag)
                for img in images:
                    img.getparent().remove(img)
                    img.tail = img.text = None
                    parent.insert(index, img)

            if hasattr(parent, 'remove'):
                parent.remove(tag)

        for tag in pagebreak_anchors:
            anchor = tag.attrib['id']
            del tag.attrib['id']
            if 'name' in tag.attrib:
                del tag.attrib['name']
            p = tag.getparent()
            a = p.makeelement('a')
            a.attrib['id'] = anchor
            p.insert(p.index(tag)+1, a)
            if getattr(a.getnext(), 'tag', None) in BLOCK_TAGS:
                forwardable_anchors.append(a)

        for tag in forwardable_anchors:
            block = tag.getnext()
            tag.getparent().remove(tag)

            if 'id' in block.attrib:
                tag.tail = block.text
                block.text = None
                block.insert(0, tag)
            else:
                block.attrib['id'] = tag.attrib['id']

        # WebKit fails to navigate to anchors located on <br> tags
        for br in root.xpath('/body/br[@id]'):
            br.tag = 'div'

    def get_left_whitespace(self, tag):

        def whitespace(tag):
            lm = ti = 0.0
            if tag.tag == 'p':
                ti = unit_convert('1.5em', 12, 500, 166)
            if tag.tag == 'blockquote':
                lm = unit_convert('2em', 12, 500, 166)
            lm = self.left_margins.get(tag, lm)
            ti = self.text_indents.get(tag, ti)
            try:
                lm = float(lm)
            except:
                lm = 0.0
            try:
                ti = float(ti)
            except:
                ti = 0.0
            return lm + ti

        parent = tag
        ans = 0.0
        while parent is not None:
            ans += whitespace(parent)
            parent = parent.getparent()

        return ans

    def create_opf(self, htmlfile, guide=None, root=None):
        mi = getattr(self.book_header.exth, 'mi', self.embedded_mi)
        if mi is None:
            mi = MetaInformation(self.book_header.title, [_('Unknown')])
        opf = OPFCreator(os.path.dirname(htmlfile), mi)
        if hasattr(self.book_header.exth, 'cover_offset'):
            opf.cover = f'images/{self.book_header.exth.cover_offset + 1:05}.jpg'
        elif mi.cover is not None:
            opf.cover = mi.cover
        else:
            opf.cover = f'images/{1:05}.jpg'
            if not os.path.exists(os.path.join(os.path.dirname(htmlfile),
                * opf.cover.split('/'))):
                opf.cover = None

        cover = opf.cover
        cover_copied = None
        if cover is not None:
            cover = cover.replace('/', os.sep)
            if os.path.exists(cover):
                ncover = 'images'+os.sep+'calibre_cover.jpg'
                if os.path.exists(ncover):
                    os.remove(ncover)
                shutil.copyfile(cover, ncover)
                cover_copied = os.path.abspath(ncover)
                opf.cover = ncover.replace(os.sep, '/')

        manifest = [(htmlfile, 'application/xhtml+xml'),
            (os.path.abspath('styles.css'), 'text/css')]
        bp = os.path.dirname(htmlfile)
        added = set()
        for i in getattr(self, 'image_names', []):
            path = os.path.join(bp, 'images', i)
            added.add(path)
            manifest.append((path, guess_type(path)[0] or 'image/jpeg'))
        if cover_copied is not None:
            manifest.append((cover_copied, 'image/jpeg'))

        opf.create_manifest(manifest)
        opf.create_spine([os.path.basename(htmlfile)])
        toc = None
        if guide is not None:
            opf.create_guide(guide)
            for ref in opf.guide:
                if ref.type.lower() == 'toc':
                    toc = ref.href()

        ncx_manifest_entry = None
        if toc:
            ncx_manifest_entry = 'toc.ncx'
            elems = root.xpath('//*[@id="{}"]'.format(toc.partition('#')[-1]))
            tocobj = None
            if elems:
                tocobj = TOC()
                found = False
                reached = False
                for x in root.iter():
                    if x == elems[-1]:
                        reached = True
                        continue
                    if reached and x.tag == 'a':
                        href = x.get('href', '')
                        if href and re.match(r'\w+://', href) is None:
                            try:
                                text = ' '.join([t.strip() for t in
                                    x.xpath('descendant::text()')])
                            except:
                                text = ''
                            text = replace_entities(text)
                            item = tocobj.add_item(toc.partition('#')[0], href[1:],
                                text)
                            item.left_space = int(self.get_left_whitespace(x))
                            found = True
                    if reached and found and x.get('class', None) == 'mbp_pagebreak':
                        break
            if tocobj is not None:
                tocobj = self.structure_toc(tocobj)
                opf.set_toc(tocobj)

        return opf, ncx_manifest_entry

    def structure_toc(self, toc):
        indent_vals = set()
        for item in toc:
            indent_vals.add(item.left_space)
        if len(indent_vals) > 6 or len(indent_vals) < 2:
            # Too many or too few levels, give up
            return toc
        indent_vals = sorted(indent_vals)

        last_found = [None for i in indent_vals]

        newtoc = TOC()

        def find_parent(level):
            candidates = last_found[:level]
            for x in reversed(candidates):
                if x is not None:
                    return x
            return newtoc

        for item in toc:
            level = indent_vals.index(item.left_space)
            parent = find_parent(level)
            last_found[level] = parent.add_item(item.href, item.fragment,
                        item.text)

        return newtoc

    def sizeof_trailing_entries(self, data):
        def sizeof_trailing_entry(ptr, psize):
            bitpos, result = 0, 0
            while True:
                v = ord(ptr[psize-1:psize])
                result |= (v & 0x7F) << bitpos
                bitpos += 7
                psize -= 1
                if (v & 0x80) != 0 or (bitpos >= 28) or (psize == 0):
                    return result

        num = 0
        size = len(data)
        flags = self.book_header.extra_flags >> 1
        while flags:
            if flags & 1:
                try:
                    num += sizeof_trailing_entry(data, size - num)
                except (IndexError, TypeError):
                    self.warn_about_trailing_entry_corruption()
                    return 0
            flags >>= 1
        if self.book_header.extra_flags & 1:
            off = size - num - 1
            try:
                num += (ord(data[off:off+1]) & 0x3) + 1
            except TypeError:
                self.log.warn('Invalid sizeof trailing entries')
                num += 1
        return num

    def warn_about_trailing_entry_corruption(self):
        if not self.warned_about_trailing_entry_corruption:
            self.warned_about_trailing_entry_corruption = True
            self.log.warn('The trailing data entries in this MOBI file are corrupted, you might see corrupted text in the output')

    def text_section(self, index):
        data = self.sections[index][0]
        trail_size = self.sizeof_trailing_entries(data)
        return data[:len(data)-trail_size]

    def extract_text(self, offset=1):
        self.log.debug('Extracting text...')
        text_sections = [self.text_section(i) for i in range(offset,
            min(self.book_header.records + offset, len(self.sections)))]
        processed_records = list(range(offset-1, self.book_header.records +
            offset))

        self.mobi_html = b''

        if self.book_header.compression_type == b'DH':
            huffs = [self.sections[i][0] for i in
                range(self.book_header.huff_offset,
                    self.book_header.huff_offset + self.book_header.huff_number)]
            processed_records += list(range(self.book_header.huff_offset,
                self.book_header.huff_offset + self.book_header.huff_number))
            huff = HuffReader(huffs)
            unpack = huff.unpack

        elif self.book_header.compression_type == b'\x00\x02':
            unpack = decompress_doc

        elif self.book_header.compression_type == b'\x00\x01':
            def unpack(x):
                return x
        else:
            raise MobiError(f'Unknown compression algorithm: {self.book_header.compression_type!r}')
        self.mobi_html = b''.join(map(unpack, text_sections))
        if self.mobi_html.endswith(b'#'):
            self.mobi_html = self.mobi_html[:-1]

        if self.book_header.ancient and b'<html' not in self.mobi_html[:300].lower():
            self.mobi_html = self.mobi_html.replace(b'\r ', b'\n\n ')
        self.mobi_html = self.mobi_html.replace(b'\0', b'')
        if self.book_header.codec == 'cp1252':
            self.mobi_html = self.mobi_html.replace(b'\x1e', b'')  # record separator
            self.mobi_html = self.mobi_html.replace(b'\x02', b'')  # start of text
        return processed_records

    def replace_page_breaks(self):
        self.processed_html = self.PAGE_BREAK_PAT.sub(
            r'<div \1 class="mbp_pagebreak" />',
            self.processed_html)

    def add_anchors(self):
        self.log.debug('Adding anchors...')
        positions = set()
        link_pattern = re.compile(br'''<[^<>]+filepos=['"]{0,1}(\d+)[^<>]*>''',
            re.IGNORECASE)
        for match in link_pattern.finditer(self.mobi_html):
            positions.add(int(match.group(1)))
        pos = 0
        processed_html = []
        end_tag_re = re.compile(br'<\s*/')
        for end in sorted(positions):
            if end == 0:
                continue
            oend = end
            l = self.mobi_html.find(b'<', end)
            r = self.mobi_html.find(b'>', end)
            anchor = b'<a id="filepos%d"></a>'
            if r > -1 and (r < l or l == end or l == -1):
                p = self.mobi_html.rfind(b'<', 0, end + 1)
                if (pos < end and p > -1 and not end_tag_re.match(self.mobi_html[p:r]) and
                        not self.mobi_html[p:r + 1].endswith(b'/>')):
                    anchor = b' filepos-id="filepos%d"'
                    end = r
                else:
                    end = r + 1
            processed_html.append(self.mobi_html[pos:end] + (anchor % oend))
            pos = end
        processed_html.append(self.mobi_html[pos:])
        processed_html = b''.join(processed_html)

        # Remove anchors placed inside entities
        self.processed_html = re.sub(br'&([^;]*?)(<a id="filepos\d+"></a>)([^;]*);',
                br'&\1\3;\2', processed_html)

    def extract_images(self, processed_records, output_dir):
        self.log.debug('Extracting images...')
        output_dir = os.path.abspath(os.path.join(output_dir, 'images'))
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        image_index = 0
        self.image_names = []
        image_name_map = {}
        start = getattr(self.book_header, 'first_image_index', -1)
        if start > self.num_sections or start < 0:
            # BAEN PRC files have bad headers
            start = 0
        for i in range(start, self.num_sections):
            if i in processed_records:
                continue
            processed_records.append(i)
            data  = self.sections[i][0]
            image_index += 1
            if data[:4] in {b'FLIS', b'FCIS', b'SRCS', b'\xe9\x8e\r\n',
                    b'RESC', b'BOUN', b'FDST', b'DATP', b'AUDI', b'VIDE'}:
                # This record is a known non image type, no need to try to
                # load the image
                continue

            try:
                imgfmt = what(None, data)
            except Exception:
                continue
            if imgfmt not in {'jpg', 'jpeg', 'gif', 'png', 'bmp'}:
                continue
            if imgfmt == 'jpeg':
                imgfmt = 'jpg'
            if imgfmt == 'gif':
                try:
                    data = gif_data_to_png_data(data)
                    imgfmt = 'png'
                except AnimatedGIF:
                    pass
                except OSError:
                    self.log.warn(f'Ignoring undecodeable GIF image at index {image_index}')
                    continue
            path = os.path.join(output_dir, f'{image_index:05}.{imgfmt}')
            image_name_map[image_index] = os.path.basename(path)
            if imgfmt == 'png':
                with open(path, 'wb') as f:
                    f.write(data)
            else:
                try:
                    save_cover_data_to(data, path, minify_to=(10000, 10000))
                except Exception:
                    continue
            self.image_names.append(os.path.basename(path))
        return image_name_map


def test_mbp_regex():
    for raw, m in iteritems({
        '<mbp:pagebreak></mbp:pagebreak>':'',
        '<mbp:pagebreak xxx></mbp:pagebreak>yyy':' xxxyyy',
        '<mbp:pagebreak> </mbp:pagebreak>':'',
        '<mbp:pagebreak>xxx':'xxx',
        '<mbp:pagebreak/>xxx':'xxx',
        '<mbp:pagebreak sdf/ >xxx':' sdfxxx',
        '<mbp:pagebreak / >':' ',
        '</mbp:pagebreak>':'',
        '</mbp:pagebreak sdf>':' sdf',
        '</mbp:pagebreak><mbp:pagebreak></mbp:pagebreak>xxx':'xxx',
        }):
        ans = MobiReader.PAGE_BREAK_PAT.sub(r'\1', raw)
        if ans != m:
            raise Exception(f'{ans!r} != {m!r} for {raw!r}')



# Form implementation generated from reading ui file '/home/kovid/work/calibre/src/calibre/gui2/convert/pdf_output.ui'
#
# Created by: PyQt6 UI code generator 6.5.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(638, 731)
        self.formLayout = QtWidgets.QFormLayout(Form)
        self.formLayout.setObjectName("formLayout")
        self.opt_use_profile_size = QtWidgets.QCheckBox(parent=Form)
        self.opt_use_profile_size.setObjectName("opt_use_profile_size")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.opt_use_profile_size)
        self.label = QtWidgets.QLabel(parent=Form)
        self.label.setObjectName("label")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label)
        self.opt_paper_size = PaperSizes(parent=Form)
        self.opt_paper_size.setObjectName("opt_paper_size")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_paper_size)
        self.label_3 = QtWidgets.QLabel(parent=Form)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_3)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.opt_custom_size = QtWidgets.QLineEdit(parent=Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.opt_custom_size.sizePolicy().hasHeightForWidth())
        self.opt_custom_size.setSizePolicy(sizePolicy)
        self.opt_custom_size.setClearButtonEnabled(True)
        self.opt_custom_size.setObjectName("opt_custom_size")
        self.horizontalLayout.addWidget(self.opt_custom_size)
        self.label_11 = QtWidgets.QLabel(parent=Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_11.sizePolicy().hasHeightForWidth())
        self.label_11.setSizePolicy(sizePolicy)
        self.label_11.setObjectName("label_11")
        self.horizontalLayout.addWidget(self.label_11)
        self.opt_unit = QtWidgets.QComboBox(parent=Form)
        self.opt_unit.setObjectName("opt_unit")
        self.horizontalLayout.addWidget(self.opt_unit)
        self.formLayout.setLayout(2, QtWidgets.QFormLayout.ItemRole.FieldRole, self.horizontalLayout)
        self.opt_preserve_cover_aspect_ratio = QtWidgets.QCheckBox(parent=Form)
        self.opt_preserve_cover_aspect_ratio.setObjectName("opt_preserve_cover_aspect_ratio")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.opt_preserve_cover_aspect_ratio)
        self.opt_pdf_page_numbers = QtWidgets.QCheckBox(parent=Form)
        self.opt_pdf_page_numbers.setObjectName("opt_pdf_page_numbers")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.opt_pdf_page_numbers)
        self.opt_pdf_add_toc = QtWidgets.QCheckBox(parent=Form)
        self.opt_pdf_add_toc.setObjectName("opt_pdf_add_toc")
        self.formLayout.setWidget(8, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.opt_pdf_add_toc)
        self.label_14 = QtWidgets.QLabel(parent=Form)
        self.label_14.setObjectName("label_14")
        self.formLayout.setWidget(9, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_14)
        self.opt_toc_title = QtWidgets.QLineEdit(parent=Form)
        self.opt_toc_title.setClearButtonEnabled(True)
        self.opt_toc_title.setObjectName("opt_toc_title")
        self.formLayout.setWidget(9, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_toc_title)
        self.label_4 = QtWidgets.QLabel(parent=Form)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(11, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_4)
        self.opt_pdf_serif_family = QtWidgets.QFontComboBox(parent=Form)
        self.opt_pdf_serif_family.setObjectName("opt_pdf_serif_family")
        self.formLayout.setWidget(11, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_pdf_serif_family)
        self.label_5 = QtWidgets.QLabel(parent=Form)
        self.label_5.setObjectName("label_5")
        self.formLayout.setWidget(12, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_5)
        self.opt_pdf_sans_family = QtWidgets.QFontComboBox(parent=Form)
        self.opt_pdf_sans_family.setObjectName("opt_pdf_sans_family")
        self.formLayout.setWidget(12, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_pdf_sans_family)
        self.label_6 = QtWidgets.QLabel(parent=Form)
        self.label_6.setObjectName("label_6")
        self.formLayout.setWidget(13, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_6)
        self.opt_pdf_mono_family = QtWidgets.QFontComboBox(parent=Form)
        self.opt_pdf_mono_family.setObjectName("opt_pdf_mono_family")
        self.formLayout.setWidget(13, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_pdf_mono_family)
        self.label_7 = QtWidgets.QLabel(parent=Form)
        self.label_7.setObjectName("label_7")
        self.formLayout.setWidget(14, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_7)
        self.opt_pdf_standard_font = QtWidgets.QComboBox(parent=Form)
        self.opt_pdf_standard_font.setObjectName("opt_pdf_standard_font")
        self.formLayout.setWidget(14, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_pdf_standard_font)
        self.label_8 = QtWidgets.QLabel(parent=Form)
        self.label_8.setObjectName("label_8")
        self.formLayout.setWidget(15, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_8)
        self.opt_pdf_default_font_size = QtWidgets.QSpinBox(parent=Form)
        self.opt_pdf_default_font_size.setObjectName("opt_pdf_default_font_size")
        self.formLayout.setWidget(15, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_pdf_default_font_size)
        self.label_9 = QtWidgets.QLabel(parent=Form)
        self.label_9.setObjectName("label_9")
        self.formLayout.setWidget(16, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_9)
        self.opt_pdf_mono_font_size = QtWidgets.QSpinBox(parent=Form)
        self.opt_pdf_mono_font_size.setObjectName("opt_pdf_mono_font_size")
        self.formLayout.setWidget(16, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_pdf_mono_font_size)
        self.label_91 = QtWidgets.QLabel(parent=Form)
        self.label_91.setObjectName("label_91")
        self.formLayout.setWidget(17, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_91)
        self.opt_pdf_page_number_map = QtWidgets.QLineEdit(parent=Form)
        self.opt_pdf_page_number_map.setClearButtonEnabled(True)
        self.opt_pdf_page_number_map.setObjectName("opt_pdf_page_number_map")
        self.formLayout.setWidget(17, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_pdf_page_number_map)
        self.page_margins_box = QtWidgets.QGroupBox(parent=Form)
        self.page_margins_box.setObjectName("page_margins_box")
        self.formLayout.setWidget(19, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.page_margins_box)
        self.template_box = QtWidgets.QGroupBox(parent=Form)
        self.template_box.setObjectName("template_box")
        self.formLayout_2 = QtWidgets.QFormLayout(self.template_box)
        self.formLayout_2.setObjectName("formLayout_2")
        self.hf_label = QtWidgets.QLabel(parent=self.template_box)
        self.hf_label.setWordWrap(True)
        self.hf_label.setOpenExternalLinks(True)
        self.hf_label.setObjectName("hf_label")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.hf_label)
        self.label_12 = QtWidgets.QLabel(parent=self.template_box)
        self.label_12.setObjectName("label_12")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_12)
        self.opt_pdf_header_template = QtWidgets.QLineEdit(parent=self.template_box)
        self.opt_pdf_header_template.setClearButtonEnabled(True)
        self.opt_pdf_header_template.setObjectName("opt_pdf_header_template")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_pdf_header_template)
        self.label_13 = QtWidgets.QLabel(parent=self.template_box)
        self.label_13.setObjectName("label_13")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_13)
        self.opt_pdf_footer_template = QtWidgets.QLineEdit(parent=self.template_box)
        self.opt_pdf_footer_template.setClearButtonEnabled(True)
        self.opt_pdf_footer_template.setObjectName("opt_pdf_footer_template")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_pdf_footer_template)
        self.formLayout.setWidget(20, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.template_box)
        self.opt_pdf_hyphenate = QtWidgets.QCheckBox(parent=Form)
        self.opt_pdf_hyphenate.setObjectName("opt_pdf_hyphenate")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.opt_pdf_hyphenate)
        self.label_2 = QtWidgets.QLabel(parent=Form)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(18, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_2)
        self.opt_pdf_odd_even_offset = QtWidgets.QDoubleSpinBox(parent=Form)
        self.opt_pdf_odd_even_offset.setDecimals(1)
        self.opt_pdf_odd_even_offset.setMinimum(-100000.0)
        self.opt_pdf_odd_even_offset.setMaximum(100000.0)
        self.opt_pdf_odd_even_offset.setObjectName("opt_pdf_odd_even_offset")
        self.formLayout.setWidget(18, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_pdf_odd_even_offset)
        self.opt_pdf_no_cover = QtWidgets.QCheckBox(parent=Form)
        self.opt_pdf_no_cover.setObjectName("opt_pdf_no_cover")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.opt_pdf_no_cover)
        self.label.setBuddy(self.opt_paper_size)
        self.label_3.setBuddy(self.opt_custom_size)
        self.label_11.setBuddy(self.opt_unit)
        self.label_14.setBuddy(self.opt_toc_title)
        self.label_4.setBuddy(self.opt_pdf_serif_family)
        self.label_5.setBuddy(self.opt_pdf_sans_family)
        self.label_6.setBuddy(self.opt_pdf_mono_family)
        self.label_7.setBuddy(self.opt_pdf_standard_font)
        self.label_8.setBuddy(self.opt_pdf_default_font_size)
        self.label_9.setBuddy(self.opt_pdf_mono_font_size)
        self.label_91.setBuddy(self.opt_pdf_page_number_map)
        self.label_12.setBuddy(self.opt_pdf_header_template)
        self.label_13.setBuddy(self.opt_pdf_footer_template)
        self.label_2.setBuddy(self.opt_pdf_odd_even_offset)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):

        Form.setWindowTitle(_("Form"))
        self.opt_use_profile_size.setText(_("&Use the paper size set in output profile"))
        self.label.setText(_("&Paper size:"))
        self.label_3.setText(_("&Custom size:"))
        self.label_11.setText(_("&Unit:"))
        self.opt_preserve_cover_aspect_ratio.setText(_("Preserve &aspect ratio of cover"))
        self.opt_pdf_page_numbers.setText(_("Add page &numbers to the bottom of every page"))
        self.opt_pdf_add_toc.setText(_("Add a printable &Table of Contents at the end"))
        self.label_14.setText(_("&Title for ToC:"))
        self.label_4.setText(_("Serif famil&y:"))
        self.label_5.setText(_("Sans fami&ly:"))
        self.label_6.setText(_("&Monospace family:"))
        self.label_7.setText(_("S&tandard font:"))
        self.label_8.setText(_("Default font si&ze:"))
        self.opt_pdf_default_font_size.setSuffix(_(" px"))
        self.label_9.setText(_("Monospace &font size:"))
        self.opt_pdf_mono_font_size.setSuffix(_(" px"))
        self.label_91.setText(_("Page &number map:"))
        self.page_margins_box.setTitle(_("Page  margins"))
        self.template_box.setTitle(_("Page headers and footers"))
        self.hf_label.setText(_("You can insert headers and footers into every page of the produced PDF file by using header and footer templates. For examples, see the <a href=\"%s\">documentation</a>."))
        self.label_12.setText(_("&Header template:"))
        self.label_13.setText(_("&Footer template:"))
        self.opt_pdf_hyphenate.setText(_("&Break long words at the end of lines"))
        self.label_2.setText(_("Odd/even &offset:"))
        self.opt_pdf_odd_even_offset.setSuffix(_(" pt"))
        self.opt_pdf_no_cover.setText(_("Discard book &cover"))
from calibre.gui2.widgets import PaperSizes
